@@ -1,9 +1,14 @@
 #ifndef INCLUDE_INJA_PARSER_HPP_
 #define INCLUDE_INJA_PARSER_HPP_
 
-#include <limits>
+#include <cstddef>
+#include <filesystem>
+#include <fstream>
+#include <iterator>
+#include <memory>
 #include <stack>
 #include <string>
+#include <string_view>
 #include <utility>
 #include <vector>
 
@@ -13,8 +18,8 @@
 #include "lexer.hpp"
 #include "node.hpp"
 #include "template.hpp"
+#include "throw.hpp"
 #include "token.hpp"
-#include "utils.hpp"
 
 namespace inja {
 
@@ -43,11 +48,11 @@ class Parser {
   std::stack<ForStatementNode*> for_statement_stack;
   std::stack<BlockStatementNode*> block_statement_stack;
 
-  inline void throw_parser_error(const std::string& message) const {
+  void throw_parser_error(const std::string& message) const {
     INJA_THROW(ParserError(message, lexer.current_position()));
   }
 
-  inline void get_next_token() {
+  void get_next_token() {
     if (have_peek_tok) {
       tok = peek_tok;
       have_peek_tok = false;
@@ -56,19 +61,19 @@ class Parser {
     }
   }
 
-  inline void get_peek_token() {
+  void get_peek_token() {
     if (!have_peek_tok) {
       peek_tok = lexer.scan();
       have_peek_tok = true;
     }
   }
 
-  inline void add_literal(Arguments &arguments, const char* content_ptr) {
-    std::string_view data_text(literal_start.data(), tok.text.data() - literal_start.data() + tok.text.size());
+  void add_literal(Arguments &arguments, const char* content_ptr) {
+    const std::string_view data_text(literal_start.data(), tok.text.data() - literal_start.data() + tok.text.size());
     arguments.emplace_back(std::make_shared<LiteralNode>(data_text, data_text.data() - content_ptr));
   }
 
-  inline void add_operator(Arguments &arguments, OperatorStack &operator_stack) {
+  void add_operator(Arguments &arguments, OperatorStack &operator_stack) {
     auto function = operator_stack.top();
     operator_stack.pop();
 
@@ -83,17 +88,16 @@ class Parser {
     arguments.emplace_back(function);
   }
 
-  void add_to_template_storage(std::string_view path, std::string& template_name) {
+  void add_to_template_storage(const std::filesystem::path& path, std::string& template_name) {
     if (template_storage.find(template_name) != template_storage.end()) {
       return;
     }
 
-    std::string original_path = static_cast<std::string>(path);
-    std::string original_name = template_name;
+    const std::string original_name = template_name;
 
     if (config.search_included_templates_in_files) {
       // Build the relative path
-      template_name = original_path + original_name;
+      template_name = (path / original_name).string();
       if (template_name.compare(0, 2, "./") == 0) {
         template_name.erase(0, 2);
       }
@@ -103,7 +107,7 @@ class Parser {
         std::ifstream file;
         file.open(template_name);
         if (!file.fail()) {
-          std::string text((std::istreambuf_iterator<char>(file)), std::istreambuf_iterator<char>());
+          const std::string text((std::istreambuf_iterator<char>(file)), std::istreambuf_iterator<char>());
 
           auto include_template = Template(text);
           template_storage.emplace(template_name, include_template);
@@ -117,7 +121,7 @@ class Parser {
 
     // Try include callback
     if (config.include_callback) {
-      auto include_template = config.include_callback(original_path, original_name);
+      auto include_template = config.include_callback(path, original_name);
       template_storage.emplace(template_name, include_template);
     }
   }
@@ -346,6 +350,47 @@ class Parser {
         }
         arguments.emplace_back(expr);
       } break;
+
+      // parse function call pipe syntax
+      case Token::Kind::Pipe: {
+        // get function name
+        get_next_token();
+        if (tok.kind != Token::Kind::Id) {
+          throw_parser_error("expected function name, got '" + tok.describe() + "'");
+        }
+        auto func = std::make_shared<FunctionNode>(tok.text, tok.text.data() - tmpl.content.c_str());
+        // add first parameter as last value from arguments
+        func->number_args += 1;
+        func->arguments.emplace_back(arguments.back());
+        arguments.pop_back();
+        get_peek_token();
+        if (peek_tok.kind == Token::Kind::LeftParen) {
+          get_next_token();
+          // parse additional parameters
+          do {
+            get_next_token();
+            auto expr = parse_expression(tmpl);
+            if (!expr) {
+              break;
+            }
+            func->number_args += 1;
+            func->arguments.emplace_back(expr);
+          } while (tok.kind == Token::Kind::Comma);
+          if (tok.kind != Token::Kind::RightParen) {
+            throw_parser_error("expected right parenthesis, got '" + tok.describe() + "'");
+          }
+        }
+        // search store for defined function with such name and number of args
+        auto function_data = function_storage.find_function(func->name, func->number_args);
+        if (function_data.operation == FunctionStorage::Operation::None) {
+          throw_parser_error("unknown function " + func->name);
+        }
+        func->operation = function_data.operation;
+        if (function_data.operation == FunctionStorage::Operation::Callback) {
+          func->callback = function_data.callback;
+        }
+        arguments.emplace_back(func);
+      } break;
       default:
         goto break_loop;
       }
@@ -368,7 +413,7 @@ class Parser {
     return expr;
   }
 
-  bool parse_statement(Template& tmpl, Token::Kind closing, std::string_view path) {
+  bool parse_statement(Template& tmpl, Token::Kind closing, const std::filesystem::path& path) {
     if (tok.kind != Token::Kind::Id) {
       return false;
     }
@@ -472,7 +517,7 @@ class Parser {
           throw_parser_error("expected id, got '" + tok.describe() + "'");
         }
 
-        Token key_token = std::move(value_token);
+        const Token key_token = value_token;
         value_token = tok;
         get_next_token();
 
@@ -533,7 +578,7 @@ class Parser {
         throw_parser_error("expected variable name, got '" + tok.describe() + "'");
       }
 
-      std::string key = static_cast<std::string>(tok.text);
+      const std::string key = static_cast<std::string>(tok.text);
       get_next_token();
 
       auto set_statement_node = std::make_shared<SetStatementNode>(key, tok.text.data() - tmpl.content.c_str());
@@ -554,7 +599,7 @@ class Parser {
     return true;
   }
 
-  void parse_into(Template& tmpl, std::string_view path) {
+  void parse_into(Template& tmpl, const std::filesystem::path& path) {
     lexer.start(tmpl.content);
     current_block = &tmpl.root;
 
@@ -569,6 +614,7 @@ class Parser {
           throw_parser_error("unmatched for");
         }
       }
+        current_block = nullptr;
         return;
       case Token::Kind::Text: {
         current_block->nodes.emplace_back(std::make_shared<TextNode>(tok.text.data() - tmpl.content.c_str(), tok.text.size()));
@@ -613,6 +659,7 @@ class Parser {
       } break;
       }
     }
+    current_block = nullptr;
   }
 
 public:
@@ -620,25 +667,22 @@ public:
                   const FunctionStorage& function_storage)
       : config(parser_config), lexer(lexer_config), template_storage(template_storage), function_storage(function_storage) {}
 
-  Template parse(std::string_view input, std::string_view path) {
-    auto result = Template(static_cast<std::string>(input));
+  Template parse(std::string_view input, const std::filesystem::path& path) {
+    auto result = Template(std::string(input));
     parse_into(result, path);
     return result;
   }
 
-  void parse_into_template(Template& tmpl, std::string_view filename) {
-    std::string_view path = filename.substr(0, filename.find_last_of("/\\") + 1);
-
-    // StringRef path = sys::path::parent_path(filename);
+  void parse_into_template(Template& tmpl, const std::filesystem::path& filename) {
     auto sub_parser = Parser(config, lexer.get_config(), template_storage, function_storage);
-    sub_parser.parse_into(tmpl, path);
+    sub_parser.parse_into(tmpl, filename.parent_path());
   }
 
-  std::string load_file(const std::string& filename) {
+  static std::string load_file(const std::filesystem::path& filename) {
     std::ifstream file;
     file.open(filename);
     if (file.fail()) {
-      INJA_THROW(FileError("failed accessing file at '" + filename + "'"));
+      INJA_THROW(FileError("failed accessing file at '" + filename.string() + "'"));
     }
     std::string text((std::istreambuf_iterator<char>(file)), std::istreambuf_iterator<char>());
     return text;
